@@ -40,6 +40,9 @@ int default_priority = 1;
 /* Save priority set by main-menu to detect priority changes from the user */
 int local_priority = -1;
 
+/* Force display of the menu at the current priority */
+int display_menu = 0;
+
 static struct debconfclient *debconf;
 
 static int di_config_package(di_system_package *p,
@@ -47,6 +50,26 @@ static int di_config_package(di_system_package *p,
 
 static void modify_debconf_priority (int raise_or_lower);
 
+static char *debconf_priorities[] =
+  {
+    "low",
+    "medium",
+    "high",
+    "critical"
+  };
+
+#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))
+static int debconf_to_pri (char *priority) {
+	int i;
+
+	if (priority)
+		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i)
+			if (0 == strcmp(priority, debconf_priorities[i]))
+				return i;
+
+	return -1;
+}
+
 /*
  * qsort comparison function (sort by menu item values, fall back to
  * lexical sort to resolve ties deterministically).
@@ -81,9 +104,19 @@ int isdefault(di_system_package *p) {
 	return false;
 }
 
+/*
+ * Menu items with a number equal to 99999 or 99900 (deprecated) are
+ * intended only to be shown during package selection in anna, but not
+ * in the main menu, so mark them uninstallable.
+ * For other packages, run the isinstallble maintainer script and check
+ * its return code.
+ */
 bool isinstallable(di_system_package *p) {
 	int check;
 
+	if ((p->installer_menu_item == 99999) || (p->installer_menu_item == 99900))
+		return false;
+
 	check = di_system_dpkg_package_control_file_exec(&p->p, "isinstallable", 0, NULL);
 	if (check <= 0)
 		return true;
@@ -139,15 +172,18 @@ get_default_menu_item(di_slist *list)
 			//di_log(DI_LOG_LEVEL_DEBUG, "not menu item; or not installed");
 			continue;
 		}
-		if ((p->installer_menu_item < last_successful_item &&
-		     !di_hash_table_lookup(seen_items, &p->p.key)) &&
-		    p->installer_menu_item < NEVERDEFAULT) {
+		if (p->installer_menu_item >= NEVERDEFAULT) {
 			//di_log(DI_LOG_LEVEL_DEBUG, "not in range to be default");
 			continue;
 		}
+		if (p->installer_menu_item < last_successful_item &&
+		    di_hash_table_lookup(seen_items, &p->p.key)) {
+			//di_log(DI_LOG_LEVEL_DEBUG, "menu item is before last_successful_item, and is not new");
+			continue;
+		}
 		/* If menutest says this item should be default, make it so */
 		if (!isdefault(p)) {
-			//di_log(DI_LOG_LEVEL_DEBUG, "isdefalt says no");
+			//di_log(DI_LOG_LEVEL_DEBUG, "isdefault says no");
 			continue;
 		}
 		/* If all of the virtual packages provided by a
@@ -183,6 +219,27 @@ static size_t menu_entry(struct debconfclient *debconf, di_system_package *packa
 	return strlen (buf);
 }
 
+/* Priority at which the menu should be displayed */
+static int menu_priority() {
+	static int default_menu_priority = -1;
+	int menu_prio = -1;
+
+	if (default_menu_priority == -1)
+		default_menu_priority = debconf_to_pri(MENU_PRIORITY);
+
+	if (display_menu)
+		menu_prio = local_priority;
+
+	if (menu_prio < 0 ||
+	    (size_t)menu_prio >= ARRAY_SIZE(debconf_priorities))
+		menu_prio = default_menu_priority;
+
+	//di_log(DI_LOG_LEVEL_INFO, "default: %i; debconf: %i; menu: %i",
+	//	default_menu_priority, local_priority, menu_prio);
+
+	return menu_prio;
+}
+
 /* Displays the main menu via debconf and returns the selected menu item. */
 di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *allocator) {
 	di_system_package **package_array, *p;
@@ -191,7 +248,7 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 	di_system_package *menudefault = NULL, *ret = NULL;
 	int i = 0, num = 0;
 	char buf[256], *menu, *s;
-	int menu_size, menu_used, size;
+	int menu_prio, menu_size, menu_used, size;
 
 	for (node = packages->list.head; node; node = node->next) {
 		p = node->data;
@@ -251,7 +308,8 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 		di_log(DI_LOG_LEVEL_INFO, "no default menu item"); 
 		modify_debconf_priority(LOWER);
 	}
-	debconf_input(debconf, MENU_PRIORITY, MAIN_MENU);
+	menu_prio = menu_priority();
+	debconf_input(debconf, debconf_priorities[menu_prio], MAIN_MENU);
 	debconf_go(debconf);
 	debconf_get(debconf, MAIN_MENU);
 	s = strdup(debconf->value);
@@ -274,6 +332,7 @@ di_system_package *show_main_menu(di_packages *packages, di_packages_allocator *
 		di_log(DI_LOG_LEVEL_WARNING, "Internal error! Cannot find \"%s\" in menu.", s);
 	}
 
+	free(s);
 	free(menu);
 	free(package_array);
 
@@ -309,8 +368,10 @@ static int satisfy_virtual(di_system_package *p) {
 			if (dep->p.status != di_package_status_installed) {
 				switch (di_config_package(dep, satisfy_virtual)) {
 					case -1:
+						di_free(menu);
 						return -1;
 					case EXIT_BACKUP:
+						di_free(menu);
 						return EXIT_BACKUP;
 				}
 			}
@@ -375,8 +436,10 @@ static int satisfy_virtual(di_system_package *p) {
 			debconf_capb(debconf, "backup");
 			debconf_subst(debconf, MISSING_PROVIDE, "CHOICES", menu);
 			debconf_input(debconf, priority, MISSING_PROVIDE);
-			if (debconf_go(debconf) != 0)
+			if (debconf_go(debconf) != 0) {
+				di_free(menu);
 				return 0;
+			}
 			debconf_capb(debconf);
 			debconf_get(debconf, MISSING_PROVIDE);
 			s = strdup(debconf->value);
@@ -392,8 +455,12 @@ static int satisfy_virtual(di_system_package *p) {
 				 * the providing packages */
 				switch (di_config_package(dep, satisfy_virtual)) {
 					case -1:
+						di_free(menu);
+						free(s);
 						return -1;
 					case EXIT_BACKUP:
+						di_free(menu);
+						free(s);
 						return EXIT_BACKUP;
 				}
 				if (is_menu_item)
@@ -401,7 +468,7 @@ static int satisfy_virtual(di_system_package *p) {
 			}
 		}
 	}
-	free(menu);
+	di_free(menu);
 	free(s);
 	/* It doesn't make sense to configure virtual packages,
 	 * since they are, well, virtual. */
@@ -426,48 +493,19 @@ static int do_menu_item(di_system_package *p) {
 	return di_config_package(p, satisfy_virtual);
 }
 
-static char *debconf_priorities[] =
-  {
-    "low",
-    "medium",
-    "high",
-    "critical"
-  };
-
-#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))
 static void modify_debconf_priority (int raise_or_lower) {
-	int pri, i;
-	static int menu_pri = -1;
+	int pri;
 	const char *template = "debconf/priority";
 	debconf_get(debconf, template);
 	
-	if (menu_pri == -1) {
-		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i) {
-			if (0 == strcmp(MENU_PRIORITY,
-					debconf_priorities[i]) ) {
-				menu_pri = i;
-				break;
-			}
-		}
-	}
+	pri = debconf_to_pri(debconf->value);
+	if ( pri == -1 )
+		pri = 1;
 
-	pri = 1;
-	if ( debconf->value ) {
-		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i) {
-			if (0 == strcmp(debconf->value,
-					debconf_priorities[i]) ) {
-				pri = i;
-				break;
-			}
-		}
-	}
 	if (raise_or_lower == LOWER) {
 		--pri;
-		/* Make sure the menu is always displayed after a single
-		 * backing up.
-		 */
-		if (menu_pri != -1 && pri > menu_pri)
-			pri = menu_pri;
+		/* Make sure the menu is always displayed after an error */
+		display_menu = 1;
 	}
 	else if (raise_or_lower == RAISE)
 		++pri;
@@ -487,22 +525,13 @@ static void modify_debconf_priority (int raise_or_lower) {
 }
 	
 static void adjust_default_priority (void) {
-	int pri, i;
+	int pri;
 	const char *template = "debconf/priority";
-	debconf_get(debconf,  template);
+	debconf_get(debconf, template);
 
-	pri = 1;
-	if ( debconf->value ) {
-		for (i = 0; (size_t)i < ARRAY_SIZE(debconf_priorities); ++i) {
-			if (0 == strcmp(debconf->value,
-					debconf_priorities[i]) ) {
-				pri = i;
-				break;
-			}
-		}
-	}
+	pri = debconf_to_pri(debconf->value);
 	
-	if ( pri != local_priority) {
+	if (pri > -1 && pri != local_priority) {
 		if (local_priority > -1)
 			di_log(DI_LOG_LEVEL_INFO, "Priority changed externally, setting main-menu default to '%s' (%s)",
 				debconf_priorities[pri] ? debconf_priorities[pri] : "(null)", debconf->value);
@@ -514,9 +543,12 @@ static void adjust_default_priority (void) {
 static void restore_default_priority (void) {
 	const char *template = "debconf/priority";
 
-	di_log(DI_LOG_LEVEL_INFO, "Restoring default debconf priority '%s'",
-		debconf_priorities[default_priority] ? debconf_priorities[default_priority] : "(null)");
-	debconf_set(debconf, template, debconf_priorities[default_priority]);
+	if (local_priority != default_priority) {
+		di_log(DI_LOG_LEVEL_INFO, "Restoring default debconf priority '%s'",
+			debconf_priorities[default_priority] ? debconf_priorities[default_priority] : "(null)");
+		debconf_set(debconf, template, debconf_priorities[default_priority]);
+		local_priority = default_priority;
+	}
 }
 
 void notify_user_of_failure (di_system_package *p) {
@@ -589,7 +621,6 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 	di_packages *packages;
 	di_packages_allocator *allocator;
 	int ret;
-	int last_item_backup = 0;
 
 	debconf = debconfclient_new();
 	di_system_init(basename(argv[0]));
@@ -603,6 +634,9 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 	    debconf->value && *debconf->value)
 		debconf_set(debconf, template, debconf->value);
 
+	/* Initialize internal priority variables */
+	adjust_default_priority();
+
 	menu_startup();
 
 	seen_items = di_hash_table_new_full(di_rstring_hash, di_rstring_equal,
@@ -613,9 +647,8 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 	while ((p=show_main_menu(packages, allocator))) {
 		di_slist_node *node;
 
-		if (p->installer_menu_item < NEVERDEFAULT && last_item_backup) {
-			restore_default_priority();
-			last_item_backup = 0;
+		if (p->installer_menu_item < NEVERDEFAULT && display_menu) {
+			display_menu = 0;
 		}
 		ret = do_menu_item(p);
 		adjust_default_priority();
@@ -624,7 +657,7 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 				/* Success */
 				if (p->installer_menu_item < NEVERDEFAULT) {
 					last_successful_item = p->installer_menu_item;
-					modify_debconf_priority(RAISE);
+					restore_default_priority();
 					//di_log(DI_LOG_LEVEL_DEBUG, "Installed package '%s', raising last_successful_item to %d", p->p.package, p->installer_menu_item);
 				}
 				else {
@@ -633,8 +666,7 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 				break;
 			case EXIT_BACKUP:
 				di_log(DI_LOG_LEVEL_INFO, "Menu item '%s' succeeded but requested to be left unconfigured.", p->p.package); 
-				modify_debconf_priority(LOWER);
-				last_item_backup = 1;
+				display_menu = 1;
 				break;
 			default:
 				di_log(DI_LOG_LEVEL_WARNING, "Menu item '%s' failed.", p->p.package);
@@ -664,6 +696,8 @@ int main (int argc __attribute__ ((unused)), char **argv) {
 	return EXIT_FAILURE;
 }
 
+int di_config_package_depth=0;
+
 /*
  * Configure all dependencies, special case for virtual packages.
  * This is done depth-first.
@@ -675,10 +709,10 @@ static int di_config_package(di_system_package *p,
 	di_slist_node *node;
 	di_system_package *dep;
 
-	//di_log(DI_LOG_LEVEL_DEBUG, "configure %s, status: %d\n", p->p.package, p->p.status);
+	//di_log(DI_LOG_LEVEL_DEBUG, "configure %s, status: %d", p->p.package, p->p.status);
 
 	if (p->p.type == di_package_type_virtual_package) {
-		//di_log(DI_LOG_LEVEL_DEBUG, "virtual package %s\n", p->p.package);
+		//di_log(DI_LOG_LEVEL_DEBUG, "virtual package %s", p->p.package);
 		if (virtfunc)
 			return virtfunc(p);
 		else
@@ -695,7 +729,14 @@ static int di_config_package(di_system_package *p,
 		if (d->type != di_package_dependency_type_depends)
 			continue;
 		/* Recursively configure this package */
-		switch (di_config_package(dep, virtfunc)) {
+		di_config_package_depth++;
+		if (di_config_package_depth > 1000) {
+			di_log(DI_LOG_LEVEL_WARNING, "Deep recursion configuring package %s (dep loop?)", p->p.package);
+			return -1;
+		}
+		ret = di_config_package(dep, virtfunc);
+		di_config_package_depth--;
+		switch (ret) {
 			case -1:
 				return -1;
 			case EXIT_BACKUP:
